<link rel="stylesheet" href="./sly/css/app/widgets/card/style.css" />

<sly-card
	{%
	if
	widget._remove_padding
	%}
	class="clear-paddings"
	{%
	endif
	%}
	:title="data.{{{widget.widget_id}}}.title"
	:subtitle="data.{{{widget.widget_id}}}.description"
	:options="data.{{{widget.widget_id}}}.options"
	:disabled="state.{{{widget.widget_id}}}.disabled"
	:collapsed="state.{{{widget.widget_id}}}.collapsed"
	style="width: 100%; {{{widget._style}}}"
>
	<div>{{{widget._content}}}</div>
	<div v-if="data.{{{widget.widget_id}}}.show_slot" slot="header">
		{{{widget._slot_content}}}
	</div>
</sly-card>
